@@ -355,7 +355,13 @@ HNTrieContainer.prototype = {
         return trieRef;
     },
 
-    serialize: function() {
+    serialize: function(encoder) {
+        if ( encoder instanceof Object ) {
+            return encoder.encode(
+                this.buf32.buffer,
+                this.buf32[HNTRIE_CHAR1_SLOT]
+            );
+        }
         return Array.from(
             new Uint32Array(
                 this.buf32.buffer,
@@ -365,24 +371,31 @@ HNTrieContainer.prototype = {
         );
     },
 
-    unserialize: function(selfie) {
-        const len = (selfie.length << 2) + HNTRIE_PAGE_SIZE-1 & ~(HNTRIE_PAGE_SIZE-1);
+    unserialize: function(selfie, decoder) {
+        this.needle = '';
+        const shouldDecode = typeof selfie === 'string';
+        let byteLength = shouldDecode
+            ? decoder.decodeSize(selfie)
+            : selfie.length << 2;
+        byteLength = byteLength + HNTRIE_PAGE_SIZE-1 & ~(HNTRIE_PAGE_SIZE-1);
+        if ( byteLength === 0 ) { return; }
         if ( this.wasmMemory !== null ) {
             const pageCountBefore = this.buf.length >>> 16;
-            const pageCountAfter = len >>> 16;
+            const pageCountAfter = byteLength >>> 16;
             if ( pageCountAfter > pageCountBefore ) {
                 this.wasmMemory.grow(pageCountAfter - pageCountBefore);
                 this.buf = new Uint8Array(this.wasmMemory.buffer);
                 this.buf32 = new Uint32Array(this.buf.buffer);
             }
+        } else if ( byteLength > this.buf.length ) {
+            this.buf = new Uint8Array(byteLength);
+            this.buf32 = new Uint32Array(this.buf.buffer);
+        }
+        if ( shouldDecode ) {
+            decoder.decode(selfie, this.buf.buffer);
         } else {
-            if ( len > this.buf.length ) {
-                this.buf = new Uint8Array(len);
-                this.buf32 = new Uint32Array(this.buf.buffer);
-            }
+            this.buf32.set(selfie);
         }
-        this.buf32.set(selfie);
-        this.needle = '';
     },
 
     //--------------------------------------------------------------------------
@@ -684,6 +697,6 @@ HNTrieContainer.prototype.HNTrieRef.prototype = {
         WebAssembly.compileStreaming
     ).catch(reason => {
         HNTrieContainer.wasmModulePromise = null;
-        console.info(reason);
+        log.info(reason);
     });
 })();
