@@ -332,6 +332,8 @@
                     raw.push(':not', '(', decompile(task[1]), ')');
                     break;
                 case ':spath':
+                    raw.push(task[1]);
+                    break;
                 case ':watch-attrs':
                 case ':xpath':
                     raw.push(task[0], '(', task[1], ')');
@@ -615,7 +617,7 @@
             const pos = hostname.lastIndexOf('.', hostname.length - 3);
             domain = pos !== -1 ? hostname.slice(pos + 1) : hostname;
         } else {
-            domain = µb.URI.domainFromHostnameNoCache(hostname);
+            domain = µb.URI.domainFromHostname(hostname);
         }
         return api.makeHash(domain);
     };
@@ -741,11 +743,22 @@
             if ( rpos === -1 ) { return false; }
         }
 
+        // https://github.com/AdguardTeam/AdguardFilters/commit/4fe02d73cee6
+        //   AdGuard also uses `$?` to force inline-based style rather than
+        //   stylesheet-based style.
         // Coarse-check that the anchor is valid.
-        // `##`: l = 1
-        // `#@#`, `#$#`, `#%#`, `#?#`: l = 2
-        // `#@$#`, `#@%#`, `#@?#`: l = 3
-        if ( (rpos - lpos) > 3 ) { return false; }
+        // `##`: l === 1
+        // `#@#`, `#$#`, `#%#`, `#?#`: l === 2
+        // `#@$#`, `#@%#`, `#@?#`, `#$?#`: l === 3
+        // `#@$?#`: l === 4
+        const anchorLen = rpos - lpos;
+        if ( anchorLen > 4 ) { return false; }
+        if (
+            anchorLen > 1 &&
+            /^@?(?:\$\??|%|\?)?$/.test(raw.slice(lpos + 1, rpos)) === false
+        ) {
+            return false;
+        }
 
         // Extract the selector.
         let suffix = raw.slice(rpos + 1).trim();
@@ -763,9 +776,8 @@
         if ( cCode !== 0x23 /* '#' */ && cCode !== 0x40 /* '@' */ ) {
             // Adguard's scriptlet injection: not supported.
             if ( cCode === 0x25 /* '%' */ ) { return true; }
-            // Not a known extended filter.
-            if ( cCode !== 0x24 /* '$' */ && cCode !== 0x3F /* '?' */ ) {
-                return false;
+            if ( cCode === 0x3F /* '?' */ && anchorLen > 2 ) {
+                cCode = raw.charCodeAt(rpos - 2);
             }
             // Adguard's style injection: translate to uBO's format.
             if ( cCode === 0x24 /* '$' */ ) {
@@ -821,18 +833,30 @@
         µb.htmlFilteringEngine.fromCompiledContent(reader, options);
     };
 
-    api.toSelfie = function() {
-        return {
-            cosmetic: µb.cosmeticFilteringEngine.toSelfie(),
-            scriptlets: µb.scriptletFilteringEngine.toSelfie(),
-            html: µb.htmlFilteringEngine.toSelfie()
-        };
+    api.toSelfie = function(path) {
+        return µBlock.assets.put(
+            `${path}/main`,
+            JSON.stringify({
+                cosmetic: µb.cosmeticFilteringEngine.toSelfie(),
+                scriptlets: µb.scriptletFilteringEngine.toSelfie(),
+                html: µb.htmlFilteringEngine.toSelfie()
+            })
+        );
     };
 
-    api.fromSelfie = function(selfie) {
-        µb.cosmeticFilteringEngine.fromSelfie(selfie.cosmetic);
-        µb.scriptletFilteringEngine.fromSelfie(selfie.scriptlets);
-        µb.htmlFilteringEngine.fromSelfie(selfie.html);
+    api.fromSelfie = function(path) {
+        return µBlock.assets.get(`${path}/main`).then(details => {
+            let selfie;
+            try {
+                selfie = JSON.parse(details.content);
+            } catch (ex) {
+            }
+            if ( selfie instanceof Object === false ) { return false; }
+            µb.cosmeticFilteringEngine.fromSelfie(selfie.cosmetic);
+            µb.scriptletFilteringEngine.fromSelfie(selfie.scriptlets);
+            µb.htmlFilteringEngine.fromSelfie(selfie.html);
+            return true;
+        });
     };
 
     return api;
