@@ -32,7 +32,7 @@
     let bytesInUse;
     let countdown = 0;
 
-    let process = count => {
+    const process = count => {
         if ( typeof count === 'number' ) {
             if ( bytesInUse === undefined ) {
                 bytesInUse = 0;
@@ -50,12 +50,11 @@
         countdown += 1;
         vAPI.storage.getBytesInUse(null, process);
     }
-    if (
-        this.cacheStorage !== vAPI.storage &&
-        this.cacheStorage.getBytesInUse instanceof Function
-    ) {
+    if ( this.cacheStorage.name !== 'browser.storage.local' ) {
         countdown += 1;
-        this.cacheStorage.getBytesInUse(null, process);
+        this.assets.getBytesInUse().then(count => {
+            process(count);
+        });
     }
     if ( countdown === 0 ) {
         callback();
@@ -92,12 +91,17 @@
 /******************************************************************************/
 
 µBlock.loadHiddenSettings = function() {
+    return new Promise(resolve => {
+    // >>>> start of executor
+
     vAPI.storage.get('hiddenSettings', bin => {
-        if ( bin instanceof Object === false ) { return; }
-        let hs = bin.hiddenSettings;
+        if ( bin instanceof Object === false ) {
+            return resolve();
+        }
+        const hs = bin.hiddenSettings;
         if ( hs instanceof Object ) {
-            let hsDefault = this.hiddenSettingsDefault;
-            for ( let key in hsDefault ) {
+            const hsDefault = this.hiddenSettingsDefault;
+            for ( const key in hsDefault ) {
                 if (
                     hsDefault.hasOwnProperty(key) &&
                     hs.hasOwnProperty(key) &&
@@ -106,10 +110,21 @@
                     this.hiddenSettings[key] = hs[key];
                 }
             }
+            if ( typeof this.hiddenSettings.suspendTabsUntilReady === 'boolean' ) {
+                this.hiddenSettings.suspendTabsUntilReady =
+                    this.hiddenSettings.suspendTabsUntilReady
+                        ? 'yes'
+                        : 'unset';
+            }
         }
         if ( vAPI.localStorage.getItem('immediateHiddenSettings') === null ) {
             this.saveImmediateHiddenSettings();
         }
+        self.log.verbosity = this.hiddenSettings.consoleLogLevel;
+        resolve();
+    });
+
+    // <<<< end of executor
     });
 };
 
@@ -118,8 +133,8 @@
 // which were not modified by the user.
 
 µBlock.saveHiddenSettings = function(callback) {
-    let bin = { hiddenSettings: {} };
-    for ( let prop in this.hiddenSettings ) {
+    const bin = { hiddenSettings: {} };
+    for ( const prop in this.hiddenSettings ) {
         if (
             this.hiddenSettings.hasOwnProperty(prop) &&
             this.hiddenSettings[prop] !== this.hiddenSettingsDefault[prop]
@@ -129,6 +144,7 @@
     }
     vAPI.storage.set(bin, callback);
     this.saveImmediateHiddenSettings();
+    self.log.verbosity = this.hiddenSettings.consoleLogLevel;
 };
 
 /******************************************************************************/
@@ -186,17 +202,13 @@
     vAPI.localStorage.setItem(
         'immediateHiddenSettings',
         JSON.stringify({
-                  cacheStorageAPI: this.hiddenSettings.cacheStorageAPI,
+                  consoleLogLevel: this.hiddenSettings.consoleLogLevel,
                disableWebAssembly: this.hiddenSettings.disableWebAssembly,
             suspendTabsUntilReady: this.hiddenSettings.suspendTabsUntilReady,
-            userResourcesLocation: this.hiddenSettings.userResourcesLocation,
         })
     );
 };
 
-// Do this here to have these hidden settings loaded ASAP.
-µBlock.loadHiddenSettings();
-
 /******************************************************************************/
 
 µBlock.savePermanentFirewallRules = function() {
@@ -239,24 +251,29 @@
 
 **/
 
-µBlock.loadSelectedFilterLists = function(callback) {
-    var µb = this;
-    vAPI.storage.get('selectedFilterLists', function(bin) {
+µBlock.loadSelectedFilterLists = function() {
+    return new Promise(resolve => {
+    // >>>> start of executor
+
+    vAPI.storage.get('selectedFilterLists', bin => {
         // Select default filter lists if first-time launch.
-        if ( !bin || Array.isArray(bin.selectedFilterLists) === false ) {
-            µb.assets.metadata(function(availableLists) {
-                µb.saveSelectedFilterLists(
-                    µb.autoSelectRegionalFilterLists(availableLists)
+        if (
+            bin instanceof Object === false ||
+            Array.isArray(bin.selectedFilterLists) === false
+        ) {
+            this.assets.metadata(availableLists => {
+                this.saveSelectedFilterLists(
+                    this.autoSelectRegionalFilterLists(availableLists)
                 );
-                callback();
+                resolve();
             });
             return;
         }
-        // TODO: Removes once 1.1.15 is in widespread use.
-        // https://github.com/gorhill/uBlock/issues/3383
-        vAPI.storage.remove('remoteBlacklists');
-        µb.selectedFilterLists = bin.selectedFilterLists;
-        callback();
+        this.selectedFilterLists = bin.selectedFilterLists;
+        resolve();
+    });
+
+    // <<<< end of executor
     });
 };
 
@@ -918,17 +935,17 @@
 // https://github.com/AdguardTeam/AdguardBrowserExtension/issues/917
 
 µBlock.processDirectives = function(content) {
-    var reIf = /^!#(if|endif)\b([^\n]*)/gm,
-        parts = [],
-        beg = 0, depth = 0, discard = false;
+    const reIf = /^!#(if|endif)\b([^\n]*)/gm;
+    const parts = [];
+    let  beg = 0, depth = 0, discard = false;
     while ( beg < content.length ) {
-        var match = reIf.exec(content);
+        const match = reIf.exec(content);
         if ( match === null ) { break; }
         if ( match[1] === 'if' ) {
-            var expr = match[2].trim();
-            var target = expr.startsWith('!');
+            let expr = match[2].trim();
+            const target = expr.startsWith('!');
             if ( target ) { expr = expr.slice(1); }
-            var token = this.processDirectives.tokens.get(expr);
+            const token = this.processDirectives.tokens.get(expr);
             if (
                 depth === 0 &&
                 discard === false &&
@@ -969,79 +986,72 @@
 /******************************************************************************/
 
 µBlock.loadRedirectResources = function(updatedContent) {
-    var µb = this,
-        content = '';
+    let content = '';
 
-    var onDone = function() {
-        µb.redirectEngine.resourcesFromString(content);
+    const onDone = ( ) => {
+        this.redirectEngine.resourcesFromString(content);
     };
 
-    var onUserResourcesLoaded = function(details) {
+    const onUserResourcesLoaded = details => {
         if ( details.content !== '' ) {
             content += '\n\n' + details.content;
         }
         onDone();
     };
 
-    var onResourcesLoaded = function(details) {
+    const onResourcesLoaded = details => {
         if ( details.content !== '' ) {
             content = details.content;
         }
-        if ( µb.hiddenSettings.userResourcesLocation === 'unset' ) {
+        if ( this.hiddenSettings.userResourcesLocation === 'unset' ) {
             return onDone();
         }
-        µb.assets.fetchText(µb.hiddenSettings.userResourcesLocation, onUserResourcesLoaded);
+        this.assets.fetchText(
+            this.hiddenSettings.userResourcesLocation,
+            onUserResourcesLoaded
+        );
     };
 
     if ( typeof updatedContent === 'string' && updatedContent.length !== 0 ) {
         return onResourcesLoaded({ content: updatedContent });
     }
 
-    var onSelfieReady = function(success) {
+    this.redirectEngine.resourcesFromSelfie().then(success => {
         if ( success !== true ) {
-            µb.assets.get('ublock-resources', onResourcesLoaded);
+            this.assets.get('ublock-resources', onResourcesLoaded);
         }
-    };
-
-    µb.redirectEngine.resourcesFromSelfie(onSelfieReady);
+    });
 };
 
 /******************************************************************************/
 
 µBlock.loadPublicSuffixList = function() {
-    return new Promise(resolve => {
-    // start of executor
-    this.assets.get('compiled/' + this.pslAssetKey, details => {
-        let selfie;
-        try {
-            selfie = JSON.parse(details.content);
-        } catch (ex) {
-        }
-        if (
-            selfie instanceof Object &&
-            publicSuffixList.fromSelfie(selfie)
-        ) {
-            resolve();
-            return;
-        }
-        this.assets.get(this.pslAssetKey, details => {
+    if ( this.hiddenSettings.disableWebAssembly === false ) {
+        publicSuffixList.enableWASM();
+    }
+
+    return this.assets.get(
+        'compiled/' + this.pslAssetKey
+    ).then(details =>
+        publicSuffixList.fromSelfie(details.content, µBlock.base128)
+    ).catch(reason => {
+        console.info(reason);
+        return false;
+    }).then(success => {
+        if ( success ) { return; }
+        return this.assets.get(this.pslAssetKey, details => {
             if ( details.content !== '' ) {
                 this.compilePublicSuffixList(details.content);
             }
-            resolve();
         });
     });
-    // end of executor
-    });
 };
 
-/******************************************************************************/
-
 µBlock.compilePublicSuffixList = function(content) {
     publicSuffixList.parse(content, punycode.toASCII);
     this.assets.put(
         'compiled/' + this.pslAssetKey,
-        JSON.stringify(publicSuffixList.toSelfie())
+        publicSuffixList.toSelfie(µBlock.base128)
     );
 };
 
@@ -1052,60 +1062,76 @@
 // some set time.
 
 µBlock.selfieManager = (function() {
-    let µb = µBlock;
-    let timer = null;
+    const µb = µBlock;
+    let timer;
 
     // As of 2018-05-31:
-    // JSON.stringify-ing ourselves results in a better baseline
-    // memory usage at selfie-load time. For some reasons.
+    //   JSON.stringify-ing ourselves results in a better baseline
+    //   memory usage at selfie-load time. For some reasons.
 
-    let create = function() {
-        timer = null;
-        let selfie = JSON.stringify({
-            magic: µb.systemSettings.selfieMagic,
-            availableFilterLists: µb.availableFilterLists,
-            staticNetFilteringEngine: µb.staticNetFilteringEngine.toSelfie(),
-            redirectEngine: µb.redirectEngine.toSelfie(),
-            staticExtFilteringEngine: µb.staticExtFilteringEngine.toSelfie()
-        });
-        µb.cacheStorage.set({ selfie: selfie });
-        µb.lz4Codec.relinquish();
-    };
-
-    let load = function(callback) {
-        µb.cacheStorage.get('selfie', function(bin) {
-            if (
-                bin instanceof Object === false ||
-                typeof bin.selfie !== 'string'
-            ) {
-                return callback(false);
-            }
-            let selfie;
-            try {
-                selfie = JSON.parse(bin.selfie);
-            } catch(ex) {
-            }
-            if (
-                selfie instanceof Object === false ||
-                selfie.magic !== µb.systemSettings.selfieMagic
-            ) {
-                return callback(false);
-            }
-            µb.availableFilterLists = selfie.availableFilterLists;
-            µb.staticNetFilteringEngine.fromSelfie(selfie.staticNetFilteringEngine);
-            µb.redirectEngine.fromSelfie(selfie.redirectEngine);
-            µb.staticExtFilteringEngine.fromSelfie(selfie.staticExtFilteringEngine);
-            callback(true);
+    const create = function() {
+        Promise.all([
+            µb.assets.put(
+                'selfie/main',
+                JSON.stringify({
+                    magic: µb.systemSettings.selfieMagic,
+                    availableFilterLists: µb.availableFilterLists,
+                })
+            ),
+            µb.redirectEngine.toSelfie('selfie/redirectEngine'),
+            µb.staticExtFilteringEngine.toSelfie('selfie/staticExtFilteringEngine'),
+            µb.staticNetFilteringEngine.toSelfie('selfie/staticNetFilteringEngine'),
+        ]).then(( ) => {
+            µb.lz4Codec.relinquish();
         });
     };
 
-    let destroy = function() {
-        if ( timer !== null ) {
+    const load = function() {
+        return Promise.all([
+            µb.assets.get('selfie/main').then(details => {
+                if (
+                    details instanceof Object === false ||
+                    typeof details.content !== 'string' ||
+                    details.content === ''
+                ) {
+                    return false;
+                }
+                let selfie;
+                try {
+                    selfie = JSON.parse(details.content);
+                } catch(ex) {
+                }
+                if (
+                    selfie instanceof Object === false ||
+                    selfie.magic !== µb.systemSettings.selfieMagic
+                ) {
+                    return false;
+                }
+                µb.availableFilterLists = selfie.availableFilterLists;
+                return true;
+            }),
+            µb.redirectEngine.fromSelfie('selfie/redirectEngine'),
+            µb.staticExtFilteringEngine.fromSelfie('selfie/staticExtFilteringEngine'),
+            µb.staticNetFilteringEngine.fromSelfie('selfie/staticNetFilteringEngine'),
+        ]).then(results =>
+            results.reduce((acc, v) => acc && v, true)
+        ).catch(reason => {
+            log.info(reason);
+            return false;
+        });
+    };
+
+    const destroy = function() {
+        if ( timer !== undefined ) {
             clearTimeout(timer);
-            timer = null;
+            timer = undefined;
         }
-        µb.cacheStorage.remove('selfie');
-        timer = vAPI.setTimeout(create, µb.selfieAfter);
+        µb.cacheStorage.remove('selfie'); // TODO: obsolete, remove eventually.
+        µb.assets.remove(/^selfie\//);
+        timer = vAPI.setTimeout(( ) => {
+            timer = undefined;
+            create();
+        }, µb.hiddenSettings.selfieAfter * 60000);
     };
 
     return {
@@ -1123,16 +1149,16 @@
 // necessarily present, i.e. administrators may removed entries which
 // values are left to the user's choice.
 
-µBlock.restoreAdminSettings = function(callback) {
-    // Support for vAPI.adminStorage is optional (webext).
+µBlock.restoreAdminSettings = function() {
+    return new Promise(resolve => {
+    // >>>> start of executor
+
     if ( vAPI.adminStorage instanceof Object === false ) {
-        callback();
-        return;
+        return resolve();
     }
 
-    var onRead = function(json) {
-        var µb = µBlock;
-        var data;
+    vAPI.adminStorage.getItem('adminSettings', json => {
+        let data;
         if ( typeof json === 'string' && json !== '' ) {
             try {
                 data = JSON.parse(json);
@@ -1141,13 +1167,12 @@
             }
         }
 
-        if ( typeof data !== 'object' || data === null ) {
-            callback();
-            return;
+        if ( data instanceof Object === false ) {
+            return resolve();
         }
 
-        var bin = {};
-        var binNotEmpty = false;
+        const bin = {};
+        let binNotEmpty = false;
 
         // Allows an admin to set their own 'assets.json' file, with their own
         // set of stock assets.
@@ -1157,8 +1182,8 @@
         }
 
         if ( typeof data.userSettings === 'object' ) {
-            for ( var name in µb.userSettings ) {
-                if ( µb.userSettings.hasOwnProperty(name) === false ) {
+            for ( const name in this.userSettings ) {
+                if ( this.userSettings.hasOwnProperty(name) === false ) {
                     continue;
                 }
                 if ( data.userSettings.hasOwnProperty(name) === false ) {
@@ -1201,13 +1226,14 @@
         }
 
         if ( typeof data.userFilters === 'string' ) {
-            µb.assets.put(µb.userFiltersPath, data.userFilters);
+            this.assets.put(this.userFiltersPath, data.userFilters);
         }
 
-        callback();
-    };
+        resolve();
+    });
 
-    vAPI.adminStorage.getItem('adminSettings', onRead);
+    // <<<< end of executor
+    });
 };
 
 /******************************************************************************/
@@ -1295,6 +1321,8 @@
 
     // Compile the list while we have the raw version in memory
     if ( topic === 'after-asset-updated' ) {
+        // Skip selfie-related content.
+        if ( details.assetKey.startsWith('selfie/') ) { return; }
         var cached = typeof details.content === 'string' && details.content !== '';
         if ( this.availableFilterLists.hasOwnProperty(details.assetKey) ) {
             if ( cached ) {
@@ -1330,8 +1358,8 @@
             cached: cached
         });
         // https://github.com/gorhill/uBlock/issues/2585
-        // Whenever an asset is overwritten, the current selfie is quite
-        // likely no longer valid.
+        //   Whenever an asset is overwritten, the current selfie is quite
+        //   likely no longer valid.
         this.selfieManager.destroy();
         return;
     }
